
from collections import defaultdict

from angr import Analysis, AnalysesHub
from angr.analyses.reaching_definitions.definition import Definition
from angr.analyses.reaching_definitions.constants import OP_AFTER

from ..block import Block
from ..statement import Assignment, Store, Call


class Simplifier(Analysis):
    """
    Perform function-level simplifications.
    """
    def __init__(self, func, func_graph=None, reaching_definitions=None):
        self.func = func
        self.func_graph = func_graph if func_graph is not None else func.graph
        self._reaching_definitions = reaching_definitions

        self.blocks = {}  # Mapping nodes to simplified blocks

        self._simplify()

    def _simplify(self):

        self._remove_dead_assignments()

    def _remove_dead_assignments(self):

        stmts_to_remove_per_block = defaultdict(set)

        # Find all statements that should be removed
        for block in self.func_graph.nodes():
            if not isinstance(block, Block):
                # Skip all blocks that are not AIL blocks
                continue
            if not block.statements:
                # Skip all empty blocks
                continue
            rd = self._reaching_definitions.get_reaching_definitions_by_node(block.addr, OP_AFTER)
            if rd is None:
                continue
            dead_defs = rd._dead_virgin_definitions

            for dead_def in dead_defs:  # type: Definition
                if dead_def.dummy:
                    continue
                stmts_to_remove_per_block[dead_def.codeloc.block_addr].add(dead_def.codeloc.stmt_idx)

        # Remove the statements
        for block in self.func_graph.nodes():
            if not isinstance(block, Block):
                continue

            if block.addr not in stmts_to_remove_per_block:
                continue

            new_statements = [ ]
            stmts_to_remove = stmts_to_remove_per_block[block.addr]

            for idx, stmt in enumerate(block.statements):
                if idx in stmts_to_remove:
                    if isinstance(stmt, (Assignment, Store)):
                        # Skip Assignment and Store statements
                        continue
                    elif isinstance(stmt, Call):
                        # the return expr is not used. it should not have return expr
                        stmt = stmt.copy()
                        stmt.ret_expr = None
                    else:
                        # Should not happen!
                        raise NotImplementedError()

                new_statements.append(stmt)

            new_block = block.copy()
            new_block.statements = new_statements

            self.blocks[block] = new_block


AnalysesHub.register_default("AILSimplifier", Simplifier)




import logging
from collections import defaultdict

from .. import Stmt, Expr, Block

from angr.engines.light import SimEngineLightVEX, SimEngineLightAIL, SpOffset, RegisterOffset
from angr import Analysis, register_analysis
from angr.analyses.forward_analysis import ForwardAnalysis, FunctionGraphVisitor, SingleNodeGraphVisitor
from angr.utils.constants import is_alignment_mask


l = logging.getLogger('ailment.analyses.propagator')


class PropagatorState:
    def __init__(self, arch):

        self.arch = arch

        self._replacements = { }
        self._final_replacements = [ ]

    def __repr__(self):
        return "<PropagatorState>"

    def copy(self):
        rd = PropagatorState(
            self.arch,
        )

        rd._replacements = self._replacements.copy()
        rd._final_replacements = self._final_replacements[ :: ]

        return rd

    def merge(self, *others):

        state = self.copy()

        keys_to_remove = set()

        for o in others:
            for k, v in o._replacements.items():
                if k not in state._replacements:
                    state._replacements[k] = v
                else:
                    if state._replacements[k] != o._replacements[k]:
                        keys_to_remove.add(k)

        for k in keys_to_remove:
            del state._replacements[k]

        return state

    def add_replacement(self, old, new):
        if new is not None:
            self._replacements[old] = new

    def get_replacement(self, old):
        return self._replacements.get(old, None)

    def remove_replacement(self, old):
        self._replacements.pop(old, None)

    def filter_replacements(self, atom):
        keys_to_remove = set()

        for k, v in self._replacements.items():
            if isinstance(v, Expr.Expression) and (v == atom or v.has_atom(atom)):
                keys_to_remove.add(k)

        for k in keys_to_remove:
            self._replacements.pop(k)

    def add_final_replacement(self, codeloc, old, new):
        self._final_replacements.append((codeloc, old, new))


def get_engine(base_engine):
    class SimEnginePropagator(base_engine):
        def __init__(self, stack_pointer_tracker=None):
            super(SimEnginePropagator, self).__init__()
            self._stack_pointer_tracker = stack_pointer_tracker

        def _process(self, state, successors, block=None):
            super(SimEnginePropagator, self)._process(state, successors, block=block)

        #
        # VEX statement handlers
        #

        def _handle_Put(self, stmt):
            raise NotImplementedError()

        def _handle_Store(self, stmt):
            raise NotImplementedError()

        #
        # VEX expression handlers
        #

        def _handle_Get(self, expr):
            raise NotImplementedError()

        def _handle_Load(self, expr):
            raise NotImplementedError()

        def _handle_DirtyExpression(self, expr):
            raise NotImplementedError()

        #
        # AIL statement handlers
        #

        def _ail_handle_Assignment(self, stmt):
            """

            :param Stmt.Assignment stmt:
            :return:
            """

            src = self._expr(stmt.src)
            dst = stmt.dst

            if type(dst) is Expr.Tmp:
                new_src = self.state.get_replacement(src)
                if new_src is not None:
                    l.debug("%s = %s, replace %s with %s.", dst, src, src, new_src)
                    self.state.add_replacement(dst, new_src)
                else:
                    l.debug("Replacing %s with %s.", dst, src)
                    self.state.add_replacement(dst, src)

            elif type(dst) is Expr.Register:
                l.debug("New replacement: %s with %s", dst, src)
                self.state.add_replacement(dst, src)

                # remove previous replacements whose source contains this register
                self.state.filter_replacements(dst)
            else:
                l.warning('Unsupported type of Assignment dst %s.', type(dst).__name__)

        def _ail_handle_Store(self, stmt):
            addr = self._expr(stmt.addr)
            data = self._expr(stmt.data)

            if isinstance(addr, Expr.StackBaseOffset):
                # Storing data to a stack variable
                self.state.add_replacement(Expr.Load(None, addr, data.bits // 8, stmt.endness), data)

        def _ail_handle_Jump(self, stmt):
            target = self._expr(stmt.target)

        def _ail_handle_Call(self, stmt):
            target = self._expr(stmt.target)

            new_args = None

            if stmt.args:
                new_args = [ ]
                for arg in stmt.args:
                    new_arg = self._expr(arg)
                    replacement = self.state.get_replacement(new_arg)
                    if replacement is not None:
                        new_args.append(replacement)
                    else:
                        new_args.append(arg)

            if new_args != stmt.args:
                new_call_stmt = Stmt.Call(stmt.idx, target, calling_convention=stmt.calling_convention,
                                          prototype=stmt.prototype, args=new_args, ret_expr=stmt.ret_expr,
                                          **stmt.tags)
                self.state.add_final_replacement(self._codeloc(),
                                                 stmt,
                                                 new_call_stmt,
                                                 )

        def _ail_handle_ConditionalJump(self, stmt):
            cond = self._expr(stmt.condition)
            true_target = self._expr(stmt.true_target)
            false_target = self._expr(stmt.false_target)

            if cond is stmt.condition and \
                    true_target is stmt.true_target and \
                    false_target is stmt.false_target:
                pass
            else:
                new_jump_stmt = Stmt.ConditionalJump(stmt.idx, cond, true_target, false_target, **stmt.tags)
                self.state.add_final_replacement(self._codeloc(),
                                                 stmt,
                                                 new_jump_stmt,
                                                 )

        #
        # AIL expression handlers
        #

        def _ail_handle_Tmp(self, expr):
            new_expr = self.state.get_replacement(expr)

            if new_expr is not None:
                l.debug("Add a final replacement: %s with %s", expr, new_expr)
                self.state.add_final_replacement(self._codeloc(), expr, new_expr)
                expr = new_expr

            return expr

        def _ail_handle_Register(self, expr):
            # Special handling for SP and BP
            if self._stack_pointer_tracker is not None:
                if expr.reg_offset == self.arch.sp_offset:
                    sb_offset = self._stack_pointer_tracker.insn_sp_offset_in(self.ins_addr)
                    if sb_offset is not None:
                        new_expr = Expr.StackBaseOffset(None, self.arch.bits, sb_offset)
                        self.state.add_final_replacement(self._codeloc(), expr, new_expr)
                        return new_expr
                elif expr.reg_offset == self.arch.bp_offset:
                    sb_offset = self._stack_pointer_tracker.insn_bp_offset_in(self.ins_addr)
                    if sb_offset is not None:
                        new_expr = Expr.StackBaseOffset(None, self.arch.bits, sb_offset)
                        self.state.add_final_replacement(self._codeloc(), expr, new_expr)
                        return new_expr

            new_expr = self.state.get_replacement(expr)
            if new_expr is not None:
                l.debug("Add a final replacement: %s with %s", expr, new_expr)
                self.state.add_final_replacement(self._codeloc(), expr, new_expr)
                expr = new_expr
            return expr

        def _ail_handle_Load(self, expr):
            addr = self._expr(expr.addr)

            if addr != expr.addr:
                return Expr.Load(expr.idx, addr, expr.size, expr.endness, **expr.tags)
            return expr

        def _ail_handle_Convert(self, expr):
            operand_expr = self._expr(expr.operand)

            if type(operand_expr) is Expr.Convert:
                if expr.from_bits == operand_expr.to_bits and expr.to_bits == operand_expr.from_bits:
                    # eliminate the redundant Convert
                    return operand_expr.operand
                else:
                    return Expr.Convert(expr.idx, operand_expr.from_bits, expr.to_bits, expr.is_signed, operand_expr.operand)
            elif type(operand_expr) is Expr.Const:
                # do the conversion right away
                value = operand_expr.value
                mask = (2 ** expr.to_bits) - 1
                value &= mask
                return Expr.Const(expr.idx, operand_expr.variable, value, expr.to_bits)

            converted = Expr.Convert(expr.idx, expr.from_bits, expr.to_bits, expr.is_signed, operand_expr)
            return converted

        def _ail_handle_Const(self, expr):
            return expr

        def _ail_handle_DirtyExpression(self, expr):
            return expr

        def _ail_handle_CmpLE(self, expr):
            operand_0 = self._expr(expr.operands[0])
            operand_1 = self._expr(expr.operands[1])

            return Expr.BinaryOp(expr.idx, 'CmpLE', [ operand_0, operand_1 ])

        def _ail_handle_CmpEQ(self, expr):
            operand_0 = self._expr(expr.operands[0])
            operand_1 = self._expr(expr.operands[1])

            return Expr.BinaryOp(expr.idx, 'CmpEQ', [ operand_0, operand_1 ])

        def _ail_handle_Add(self, expr):
            operand_0 = self._expr(expr.operands[0])
            operand_1 = self._expr(expr.operands[1])

            if isinstance(operand_0, Expr.Const) and isinstance(operand_1, Expr.Const):
                return Expr.Const(expr.idx, None, operand_0.value + operand_1.value, expr.bits)
            elif isinstance(operand_0, Expr.BasePointerOffset) and isinstance(operand_1, Expr.Const):
                r = operand_0.copy()
                r.offset += operand_1.value
                return r
            return Expr.BinaryOp(expr.idx, 'Add', [operand_0 if operand_0 is not None else expr.operands[0],
                                                   operand_1 if operand_1 is not None else expr.operands[1]
                                                   ])

        def _ail_handle_Sub(self, expr):
            operand_0 = self._expr(expr.operands[0])
            operand_1 = self._expr(expr.operands[1])

            if isinstance(operand_0, Expr.Const) and isinstance(operand_1, Expr.Const):
                return Expr.Const(expr.idx, None, operand_0.value - operand_1.value, expr.bits)
            elif isinstance(operand_0, Expr.BasePointerOffset) and isinstance(operand_1, Expr.Const):
                r = operand_0.copy()
                r.offset -= operand_1.value
                return r
            return Expr.BinaryOp(expr.idx, 'Sub', [ operand_0 if operand_0 is not None else expr.operands[0],
                                                    operand_1 if operand_1 is not None else expr.operands[1]
                                                    ])

        def _ail_handle_StackBaseOffset(self, expr):
            return expr

        def _ail_handle_And(self, expr):
            operand_0 = self._expr(expr.operands[0])
            operand_1 = self._expr(expr.operands[1])

            # Special logic for SP alignment
            if type(operand_0) is Expr.StackBaseOffset and \
                    type(operand_1) is Expr.Const and is_alignment_mask(operand_1.value):
                return operand_0

            return Expr.BinaryOp(expr.idx, 'And', [ operand_0, operand_1 ])

        def _ail_handle_Xor(self, expr):
            operand_0 = self._expr(expr.operands[0])
            operand_1 = self._expr(expr.operands[1])

            return Expr.BinaryOp(expr.idx, 'Xor', [ operand_0, operand_1 ])

    return SimEnginePropagator


class Propagator(ForwardAnalysis, Analysis):
    def __init__(self, func=None, block=None, max_iterations=3, stack_pointer_tracker=None):
        """

        """

        if func is not None:
            # traversing a function
            graph_visitor = FunctionGraphVisitor(func)
        elif block is not None:
            # traversing a block
            graph_visitor = SingleNodeGraphVisitor(block)
        else:
            raise ValueError('HUH')

        ForwardAnalysis.__init__(self, order_jobs=True, allow_merging=True, allow_widening=False,
                                 graph_visitor=graph_visitor)

        self._max_iterations = max_iterations
        self._function = func
        self._block = block
        self._stack_pointer_tracker = stack_pointer_tracker

        self._node_iterations = defaultdict(int)
        self._states = { }

        self._engine_vex = get_engine(SimEngineLightVEX)(stack_pointer_tracker=self._stack_pointer_tracker)
        self._engine_ail = get_engine(SimEngineLightAIL)(stack_pointer_tracker=self._stack_pointer_tracker)

        self._analyze()

    #
    # Main analysis routines
    #

    def _pre_analysis(self):
        pass

    def _pre_job_handling(self, job):
        pass

    def _initial_abstract_state(self, node):
        return PropagatorState(self.project.arch)

    def _merge_states(self, node, *states):
        return states[0].merge(*states[1:])

    def _run_on_node(self, node, state):

        input_state = state

        if isinstance(node, Block):
            block = node
            block_key = node.addr
            engine = self._engine_ail
        else:
            block = self.project.factory.block(node.addr, node.size, opt_level=0)
            block_key = node.addr
            engine = self._engine_vex

        state = input_state.copy()

        engine.process(state, block=block)

        self._node_iterations[block_key] += 1

        self._states[block_key] = state

        return True, state

    def _intra_analysis(self):
        pass

    def _post_analysis(self):
        pass

register_analysis(Propagator, "AILPropagator")




from .block import Block
from .converter import IRSBConverter
from .manager import Manager

from . import statement as Stmt
from . import expression as Expr
from .statement import Assignment
from .expression import Expression, Const, Tmp, Register, UnaryOp, BinaryOp




from .tagged_object import TaggedObject


class Expression(TaggedObject):
    """
    The base class of all AIL expressions.
    """
    def __init__(self, idx, **kwargs):
        super(Expression, self).__init__(**kwargs)
        self.idx = idx

    def __repr__(self):
        raise NotImplementedError()

    def has_atom(self, atom):
        return False

    def replace(self, old_expr, new_expr):
        if self is old_expr:
            r = True
            replaced = new_expr
        elif not isinstance(self, Atom):
            r, replaced = self.replace(old_expr, new_expr)
        else:
            r, replaced = False, self

        return r, replaced

    def __add__(self, other):
        return BinaryOp(None, 'Add', [ self, other ])

    def __sub__(self, other):
        return BinaryOp(None, 'Sub', [ self, other ])


class Atom(Expression):
    def __init__(self, idx, variable, **kwargs):
        super(Atom, self).__init__(idx, **kwargs)
        self.variable = variable

    def __repr__(self):
        return "Atom (%d)" % self.idx


class Const(Atom):
    def __init__(self, idx, variable, value, bits, **kwargs):
        super(Const, self).__init__(idx, variable, **kwargs)

        self.value = value
        self.bits = bits

    @property
    def size(self):
        return self.bits // 8

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "%#x<%d>" % (self.value, self.bits)

    def __eq__(self, other):
        return type(self) is type(other) and \
            self.value == other.value and \
            self.bits == other.bits

    def __hash__(self):
        return hash((self.value, self.bits))

    @property
    def sign_bit(self):
        return self.value >> (self.bits - 1)


class Tmp(Atom):
    def __init__(self, idx, variable, tmp_idx, bits, **kwargs):
        super(Tmp, self).__init__(idx, variable, **kwargs)

        self.tmp_idx = tmp_idx
        self.bits = bits

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "t%d" % self.tmp_idx

    def __eq__(self, other):
        return type(self) is type(other) and \
            self.tmp_idx == other.tmp_idx and \
            self.bits == other.bits

    def __hash__(self):
        return hash(('tmp', self.tmp_idx, self.bits))


class Register(Atom):
    def __init__(self, idx, variable, reg_offset, bits, **kwargs):
        super(Register, self).__init__(idx, variable, **kwargs)

        self.reg_offset = reg_offset
        self.bits = bits

    @property
    def size(self):
        return self.bits // 8

    def __repr__(self):
        return str(self)

    def __str__(self):
        if self.variable is None:
            if hasattr(self, 'reg_name'):
                return "%s<%d>" % (self.reg_name, self.bits // 8)
            else:
                return "reg_%d<%d>" % (self.reg_offset, self.bits // 8)
        else:
            return "%s" % str(self.variable.name)

    def __eq__(self, other):
        return type(self) is type(other) and \
               self.reg_offset == other.reg_offset and \
            self.bits == other.bits

    def __hash__(self):
        return hash(('reg', self.reg_offset, self.bits))


class Op(Expression):
    def __init__(self, idx, op, **kwargs):
        super(Op, self).__init__(idx, **kwargs)
        self.op = op


class UnaryOp(Op):
    def __init__(self, idx, op, operand, **kwargs):
        super(UnaryOp, self).__init__(idx, op, **kwargs)

        self.operand = operand
        self.bits = operand.bits

    def __str__(self):
        return "(%s %s)" % (self.op, str(self.operand))

    def __repr__(self):
        return str(self)

    def replace(self, old_expr, new_expr):
        r, replaced_operand = self.operand.replace(old_expr, new_expr)

        if r:
            return True, UnaryOp(self.idx, self.op, replaced_operand, **self.tags)
        else:
            return False, self


class Convert(UnaryOp):
    def __init__(self, idx, from_bits, to_bits, is_signed, operand, **kwargs):
        super(Convert, self).__init__(idx, 'Convert', operand, **kwargs)

        self.from_bits = from_bits
        self.to_bits = to_bits
        # override the size
        self.bits = to_bits
        self.is_signed = is_signed

    def __str__(self):
        return "Conv(%d->%d, %s)" % (self.from_bits, self.to_bits, self.operand)

    def __repr__(self):
        return str(self)

    def replace(self, old_expr, new_expr):
        r, replaced_operand = self.operand.replace(old_expr, new_expr)

        if r:
            return True, Convert(self.idx, self.from_bits, self.to_bits, self.is_signed, replaced_operand, **self.tags)
        else:
            return False, self


class BinaryOp(Op):
    OPSTR_MAP = {
        'Add': '+',
        'Sub': '-',
        'CmpEQ': '==',
        'CmpNE': '!=',
        'CmpLT': '<',
        'CmpLE': '<=',
        'CmpGT': '>',
        'CmpGE': '>=',
    }

    def __init__(self, idx, op, operands, **kwargs):
        super(BinaryOp, self).__init__(idx, op, **kwargs)

        assert len(operands) == 2
        self.operands = operands
        self.bits = operands[0].bits if type(operands[0]) is not int else operands[1].bits

        # TODO: sanity check of operands' sizes for some ops
        # assert self.bits == operands[1].bits

    def __str__(self):
        op_str = self.OPSTR_MAP.get(self.op, self.op)
        return "(%s %s %s)" % (str(self.operands[0]), op_str, str(self.operands[1]))

    def __repr__(self):
        return "%s(%s, %s)" % (self.op, self.operands[0], self.operands[1])

    def __eq__(self, other):
        return type(other) is BinaryOp and \
               self.operands == other.operands and \
               self.op == other.op and \
               self.bits == other.bits

    def __hash__(self):
        return hash((self.op, tuple(self.operands), self.bits))

    def has_atom(self, atom):
        for op in self.operands:
            if op == atom or op.has_atom(atom):
                return True
        return False

    def replace(self, old_expr, new_expr):
        r0, replaced_operand_0 = self.operands[0].replace(old_expr, new_expr)
        r1, replaced_operand_1 = self.operands[1].replace(old_expr, new_expr)

        if r0 or r1:
            return True, BinaryOp(self.idx, self.op, [ replaced_operand_0, replaced_operand_1 ], **self.tags)
        else:
            return False, self


class Load(Expression):
    def __init__(self, idx, addr, size, endness, **kwargs):
        super(Load, self).__init__(idx, **kwargs)

        self.addr = addr
        self.size = size
        self.endness = endness

    @property
    def bits(self):
        return self.size * 8

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "Load(addr=%s, size=%d, endness=%s)" % (self.addr, self.size, self.endness)

    def has_atom(self, atom):
        if type(self.addr) is int:
            return False
        return self.addr.has_atom(atom)

    def replace(self, old_expr, new_expr):
        r, replaced_addr = self.addr.replace(old_expr, new_expr)

        if r:
            return True, Load(self.idx, replaced_addr, self.size, self.endness, **self.tags)
        else:
            return False, self

    def __eq__(self, other):
        return type(other) is Load and \
               self.addr == other.addr and \
               self.size == other.size and \
               self.endness == other.endness

    def __hash__(self):
        return hash(('Load', self.addr, self.size, self.endness))


class ITE(Expression):
    def __init__(self, idx, cond, iffalse, iftrue, **kwargs):
        super(ITE, self).__init__(idx, **kwargs)

        self.cond = cond
        self.iffalse = iffalse
        self.iftrue = iftrue
        self.bits = iftrue.bits

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "((%s) ? (%s) : (%s))" % (self.cond, self.iftrue, self.iffalse)

    def has_atom(self, atom):
        return self.cond.has_atom(atom) or self.iftrue.has_atom(atom) or self.iffalse.has_atom(atom)

    def replace(self, old_expr, new_expr):
        cond_replaced, new_cond = self.cond.replace(old_expr, new_expr)
        iffalse_replaced, new_iffalse = self.iffalse.replace(old_expr, new_expr)
        iftrue_replaced, new_iftrue = self.iftrue.replace(old_expr, new_expr)
        replaced = cond_replaced or iftrue_replaced or iffalse_replaced

        if replaced:
            return True, ITE(self.idx, new_cond, new_iffalse, new_iftrue, **self.tags)
        else:
            return False, self


class DirtyExpression(Expression):
    def __init__(self, idx, dirty_expr, **kwargs):
        super(DirtyExpression, self).__init__(idx, **kwargs)
        self.dirty_expr = dirty_expr

    def replace(self, old_expr, new_expr):
        return False, self

    def __repr__(self):
        return "DirtyExpression (%s)" % type(self.dirty_expr)

    def __str__(self):
        return "[D] %s" % str(self.dirty_expr)


#
# Special (Dummy) expressions
#


class BasePointerOffset(Expression):
    def __init__(self, idx, bits, base, offset, **kwargs):
        super().__init__(idx, **kwargs)
        self.bits = bits
        self.base = base
        self.offset = offset

    @property
    def size(self):
        return self.bits // 8

    def __repr__(self):
        if self.offset is None:
            return "BaseOffset(%s)" % self.base
        return "BaseOffset(%s, %d)" % (self.base, self.offset)

    def __str__(self):
        if self.offset is None:
            return str(self.base)
        return "%s%+d" % (self.base, self.offset)

    def __eq__(self, other):
        return type(other) is type(self) and \
               self.bits == other.bits and \
               self.base == other.base and \
               self.offset == other.offset

    def __hash__(self):
        return hash((self.bits, self.base, self.offset))

    def replace(self, old_expr, new_expr):
        if isinstance(self.base, Expression):
            base_replaced, new_base = self.base.replace(old_expr, new_expr)
        else:
            base_replaced, new_base = False, self.base
        if isinstance(self.offset, Expression):
            offset_replaced, new_offset = self.offset.replace(old_expr, new_expr)
        else:
            offset_replaced, new_offset = False, self.offset

        if base_replaced or offset_replaced:
            return True, BasePointerOffset(self.idx, self.bits, new_base, new_offset, **self.tags)
        return False, self

    def copy(self):
        return BasePointerOffset(self.idx, self.bits, self.base, self.offset, **self.tags)


class StackBaseOffset(BasePointerOffset):
    def __init__(self, idx, bits, offset, **kwargs):
        super().__init__(idx, bits, 'stack_base', offset, **kwargs)

    def copy(self):
        return StackBaseOffset(self.idx, self.bits, self.offset, **self.tags)




import logging

from angr import Analysis, register_analysis
from angr.analyses.reaching_definitions import OP_AFTER
from angr.analyses.reaching_definitions.external_codeloc import ExternalCodeLocation

from ..block import Block
from ..statement import Assignment
from ..expression import Tmp, Register

_l = logging.getLogger(name=__name__)


class BlockSimplifier(Analysis):
    """
    Simplify an AIL block.
    """
    def __init__(self, block, stack_pointer_tracker=None):
        """

        :param Block block:
        """

        self.block = block
        self._stack_pointer_tracker = stack_pointer_tracker

        self.result_block = None

        self._analyze()

    def _analyze(self):

        block = self.block
        ctr = 0
        max_ctr = 30

        while True:
            ctr += 1
            # print(str(block))
            new_block = self._simplify_block_once(block)
            # print()
            # print(str(new_block))
            if new_block == block:
                break
            block = new_block
            if ctr >= max_ctr:
                _l.error("Simplification does not reach a fixed point after %d iterations. "
                         "Block comparison is probably incorrect." % max_ctr)
                break

        self.result_block = block

    def _simplify_block_once(self, block):

        # propagator
        propagator = self.project.analyses.AILPropagator(block=block, stack_pointer_tracker=self._stack_pointer_tracker)
        replacements = list(propagator._states.values())[0]._final_replacements
        new_block = self._replace_and_build(block, replacements)
        new_block = self._eliminate_dead_assignments(new_block)

        return new_block

    @staticmethod
    def _replace_and_build(block, replacements):

        new_statements = block.statements[::]

        for codeloc, old, new in replacements:
            stmt = new_statements[codeloc.stmt_idx]
            if stmt == old:
                # replace this statement
                r = True
                new_stmt = new
            else:
                # replace the expressions involved in this statement
                r, new_stmt = stmt.replace(old, new)

            if r:
                new_statements[codeloc.stmt_idx] = new_stmt

        new_block = block.copy()
        new_block.statements = new_statements

        return new_block

    def _eliminate_dead_assignments(self, block):

        new_statements = [ ]
        if not block.statements:
            return block

        rd = self.project.analyses.ReachingDefinitions(block=block,
                                                       track_tmps=True,
                                                       observation_points=[('insn', block.statements[-1].ins_addr, OP_AFTER)]
                                                       )

        used_tmp_indices = set(rd.one_result.tmp_uses.keys())
        dead_virgins = rd.one_result._dead_virgin_definitions
        dead_virgins_stmt_idx = set([ d.codeloc.stmt_idx for d in dead_virgins
                                      if not isinstance(d.codeloc, ExternalCodeLocation) and not d.dummy ])

        for idx, stmt in enumerate(block.statements):
            if type(stmt) is Assignment:
                if type(stmt.dst) is Tmp:
                    if stmt.dst.tmp_idx not in used_tmp_indices:
                        continue

                # is it a dead virgin?
                if idx in dead_virgins_stmt_idx:
                    continue

                # is it an assignment to an artificial register?
                if type(stmt.dst) is Register and self.project.arch.is_artificial_register(stmt.dst.reg_offset, stmt.dst.size):
                    continue

            new_statements.append(stmt)

        new_block = block.copy()
        new_block.statements = new_statements

        return new_block


register_analysis(BlockSimplifier, 'AILBlockSimplifier')




import logging

import pyvex
from angr.engines.vex.irop import operations as vex_operations

from .block import Block
from .statement import Assignment, Store, Jump, ConditionalJump, Call, DirtyStatement
from .expression import Atom, Const, Register, Tmp, DirtyExpression, UnaryOp, Convert, BinaryOp, Load, ITE

l = logging.getLogger(name=__name__)


class SkipConversionNotice(Exception):
    pass


class Converter:
    @staticmethod
    def convert(thing):
        raise NotImplementedError()


class VEXExprConverter(Converter):

    @staticmethod
    def generic_name_from_vex_op(vex_op):
        simop = vex_operations[vex_op]
        return simop._generic_name

    @staticmethod
    def convert(expr, manager):
        """

        :param expr:
        :return:
        """
        try:
            func = EXPRESSION_MAPPINGS[type(expr)]
        except KeyError:
            l.warning("VEXExprConverter: Unsupported VEX expression of type %s.", type(expr))
            return DirtyExpression(manager.next_atom(), expr, bits=expr.result_size(manager.tyenv))

        return func(expr, manager)

    @staticmethod
    def convert_list(exprs, manager):

        converted = [ ]
        for expr in exprs:
            converted.append(VEXExprConverter.convert(expr, manager))
        return converted

    @staticmethod
    def register(offset, bits, manager):
        reg_size = bits // manager.arch.byte_width
        reg_name = manager.arch.translate_register_name(offset, reg_size)
        return Register(manager.next_atom(), None, offset, bits, reg_name=reg_name)

    @staticmethod
    def tmp(tmp_idx, bits, manager):
        return Tmp(manager.next_atom(), None, tmp_idx, bits)

    @staticmethod
    def RdTmp(expr, manager):
        return VEXExprConverter.tmp(expr.tmp, expr.result_size(manager.tyenv), manager)

    @staticmethod
    def Get(expr, manager):
        return VEXExprConverter.register(expr.offset, expr.result_size(manager.tyenv), manager)

    @staticmethod
    def Load(expr, manager):
        return Load(manager.next_atom(),
                    VEXExprConverter.convert(expr.addr, manager),
                    expr.result_size(manager.tyenv) // 8,
                    expr.end
                    )

    @staticmethod
    def Unop(expr, manager):
        op_name = VEXExprConverter.generic_name_from_vex_op(expr.op)
        if op_name is None:
            # is it a convertion?
            simop = vex_operations[expr.op]
            if simop._conversion:
                return Convert(manager.next_atom(),
                               simop._from_size,
                               simop._to_size,
                               simop.is_signed,
                               VEXExprConverter.convert(expr.args[0], manager),
                               )
            raise NotImplementedError('Unsupported operation')

        return UnaryOp(manager.next_atom(),
                       op_name,
                       VEXExprConverter.convert(expr.args[0], manager),
                       )

    @staticmethod
    def Binop(expr, manager):
        op = VEXExprConverter.generic_name_from_vex_op(expr.op)
        operands = VEXExprConverter.convert_list(expr.args, manager)

        if op == 'Add' and \
                type(operands[1]) is Const and \
                operands[1].sign_bit == 1:
            # convert it to a sub
            op = 'Sub'
            op1_val, op1_bits = operands[1].value, operands[1].bits
            operands[1] = Const(operands[1].idx, None, (1 << op1_bits) - op1_val, op1_bits)

        return BinaryOp(manager.next_atom(),
                        op,
                        operands
                        )

    @staticmethod
    def Const(expr, manager):
        # pyvex.IRExpr.Const
        return Const(manager.next_atom(), None, expr.con.value, expr.result_size(manager.tyenv))

    @staticmethod
    def const_32(expr, manager):
        # pyvex.const.xxx
        return Const(manager.next_atom(), None, expr.value, 32)

    @staticmethod
    def const_64(expr, manager):
        # pyvex.const.xxx
        return Const(manager.next_atom(), None, expr.value, 64)

    @staticmethod
    def ITE(expr, manager):
        cond = VEXExprConverter.convert(expr.cond, manager)
        iffalse = VEXExprConverter.convert(expr.iffalse, manager)
        iftrue = VEXExprConverter.convert(expr.iftrue, manager)

        return ITE(manager.next_atom(), cond, iffalse, iftrue)


EXPRESSION_MAPPINGS = {
    pyvex.IRExpr.RdTmp: VEXExprConverter.RdTmp,
    pyvex.IRExpr.Get: VEXExprConverter.Get,
    pyvex.IRExpr.Unop: VEXExprConverter.Unop,
    pyvex.IRExpr.Binop: VEXExprConverter.Binop,
    pyvex.IRExpr.Const: VEXExprConverter.Const,
    pyvex.const.U32: VEXExprConverter.const_32,
    pyvex.const.U64: VEXExprConverter.const_64,
    pyvex.IRExpr.Load: VEXExprConverter.Load,
    pyvex.IRExpr.ITE: VEXExprConverter.ITE,
}


class VEXStmtConverter(Converter):

    @staticmethod
    def convert(idx, stmt, manager):
        """

        :param idx:
        :param stmt:
        :param manager:
        :return:
        """

        try:
            func = STATEMENT_MAPPINGS[type(stmt)]
        except KeyError:
            return DirtyStatement(idx, stmt, ins_addr=manager.ins_addr)

        return func(idx, stmt, manager)

    @staticmethod
    def WrTmp(idx, stmt, manager):

        var = VEXExprConverter.tmp(stmt.tmp, stmt.data.result_size(manager.tyenv), manager)
        reg = VEXExprConverter.convert(stmt.data, manager)

        return Assignment(idx, var, reg, ins_addr=manager.ins_addr,)

    @staticmethod
    def Put(idx, stmt, manager):
        data = VEXExprConverter.convert(stmt.data, manager)
        reg = VEXExprConverter.register(stmt.offset, data.bits, manager)
        return Assignment(idx, reg, data, ins_addr=manager.ins_addr,)

    @staticmethod
    def Store(idx, stmt, manager):

        return Store(idx,
                     VEXExprConverter.convert(stmt.addr, manager),
                     VEXExprConverter.convert(stmt.data, manager),
                     stmt.data.result_size(manager.tyenv) // 8,
                     stmt.endness,
                     ins_addr=manager.ins_addr,
                     )

    @staticmethod
    def Exit(idx, stmt, manager):

        if stmt.jumpkind in {'Ijk_EmWarn', 'Ijk_NoDecode',
                              'Ijk_MapFail', 'Ijk_NoRedir',
                              'Ijk_SigTRAP', 'Ijk_SigSEGV',
                              'Ijk_ClientReq'}:
            raise SkipConversionNotice()

        return ConditionalJump(idx,
                               VEXExprConverter.convert(stmt.guard, manager),
                               VEXExprConverter.convert(stmt.dst, manager),
                               None,  # it will be filled in right afterwards
                               ins_addr=manager.ins_addr
                               )


STATEMENT_MAPPINGS = {
    pyvex.IRStmt.Put: VEXStmtConverter.Put,
    pyvex.IRStmt.WrTmp: VEXStmtConverter.WrTmp,
    pyvex.IRStmt.Store: VEXStmtConverter.Store,
    pyvex.IRStmt.Exit: VEXStmtConverter.Exit,
}


class IRSBConverter(Converter):

    @staticmethod
    def convert(irsb, manager):
        """

        :param irsb:
        :param manager:
        :return:
        """

        # convert each VEX statement into an AIL statement
        statements = [ ]
        idx = 0

        manager.tyenv = irsb.tyenv

        addr = None

        conditional_jumps = [ ]

        for stmt in irsb.statements:
            if type(stmt) is pyvex.IRStmt.IMark:
                if addr is None:
                    addr = stmt.addr + stmt.delta
                manager.ins_addr = stmt.addr + stmt.delta
                continue
            elif type(stmt) is pyvex.IRStmt.AbiHint:
                # TODO: How can we use AbiHint?
                continue

            try:
                converted = VEXStmtConverter.convert(idx, stmt, manager)
                statements.append(converted)
                if type(converted) is ConditionalJump:
                    conditional_jumps.append(converted)
            except SkipConversionNotice:
                pass

            idx += 1

        if irsb.jumpkind == 'Ijk_Call':
            # call

            # TODO: is there a conditional call?

            ret_reg_offset = manager.arch.ret_offset
            ret_expr = Register(None, None, ret_reg_offset, manager.arch.bits)

            statements.append(Call(manager.next_atom(),
                                   VEXExprConverter.convert(irsb.next, manager),
                                   ret_expr=ret_expr,
                                   ins_addr=manager.ins_addr
                                   )
                              )
        elif irsb.jumpkind == 'Ijk_Boring':
            if len(conditional_jumps) == 1:
                # fill in the false target
                cond_jump = conditional_jumps[0]
                cond_jump.false_target = VEXExprConverter.convert(irsb.next, manager)

            else:
                # jump
                statements.append(Jump(manager.next_atom(),
                                       VEXExprConverter.convert(irsb.next, manager),
                                       ins_addr=manager.ins_addr
                                       )
                                  )

        return Block(addr, irsb.size, statements=statements)
